# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

from textwrap import dedent
from unittest import SkipTest

import pytest
from django.core.management import CommandError, call_command
from django.db import connection
from django.db.utils import ConnectionHandler
from django.test import SimpleTestCase, TestCase, TransactionTestCase
from django.utils.six.moves import StringIO

from django_mysql.management.commands.fix_datetime_columns import (
    parse_create_table,
)
from django_mysql.utils import connection_is_mariadb

try:
    from unittest import mock
except ImportError:
    import mock

# Can't use @override_settings to swap out DATABASES, instead just mock.patch
# a new ConnectionHandler into the command module

command_connections = (
    'django_mysql.management.commands.fix_datetime_columns.connections'
)

sqlite = ConnectionHandler({
    'default': {'ENGINE': 'django.db.backends.sqlite3'},
})


def run_it(*args, **kwargs):
    run_args = ['fix_datetime_columns']
    run_args.extend(args)

    out = StringIO()
    run_kwargs = {'stdout': out, 'skip_checks': True}
    run_kwargs.update(kwargs)

    call_command(*run_args, **run_kwargs)

    return out.getvalue()


class Datetime6TestMixin(object):

    @classmethod
    def setUpClass(cls):
        if (
            connection_is_mariadb(connection)
            or connection.mysql_version[:2] < (5, 6)
        ):
            raise SkipTest(
                "Django only uses datetime(6) columns on MySQL 5.6+",
            )
        super(Datetime6TestMixin, cls).setUpClass()


class FixDatetimeColumnsTests(Datetime6TestMixin, TestCase):

    multi_db = True

    def test_nothing_by_default(self):
        assert run_it() == ''

    def test_nothing_by_default_alternative_connection(self):
        assert run_it('other') == ''

    def test_invalid_database(self):
        with pytest.raises(CommandError) as excinfo:
            run_it('bla')

        assert "does not exist" in str(excinfo.value)

    @mock.patch(command_connections, sqlite)
    def test_invalid_not_mysql(self):
        with pytest.raises(CommandError) as excinfo:
            run_it()
        assert "not a MySQL database connection" in str(excinfo.value)


class SlowFixDatetimeColumnsTests(Datetime6TestMixin, TransactionTestCase):

    def test_with_one_column_to_migrate(self):
        with connection.cursor() as cursor:
            cursor.execute("""
                ALTER TABLE testapp_author
                    MODIFY COLUMN birthday datetime DEFAULT NULL
            """)
            try:
                out = run_it()
            finally:
                cursor.execute("""
                    ALTER TABLE testapp_author
                        MODIFY COLUMN birthday datetime(6) DEFAULT NULL
                """)
        assert out == dedent('''\
            ALTER TABLE `testapp_author`
                MODIFY COLUMN `birthday` datetime(6) DEFAULT NULL;
            ''')

    def test_with_two_columns_to_migrate(self):
        with connection.cursor() as cursor:
            cursor.execute("""
                ALTER TABLE testapp_author
                    MODIFY COLUMN birthday datetime DEFAULT NULL,
                    MODIFY COLUMN deathday datetime DEFAULT NULL
            """)
            try:
                out = run_it()
            finally:
                cursor.execute("""
                    ALTER TABLE testapp_author
                        MODIFY COLUMN birthday datetime(6) DEFAULT NULL,
                        MODIFY COLUMN deathday datetime(6) DEFAULT NULL
                """)
        assert out == dedent('''\
            ALTER TABLE `testapp_author`
                MODIFY COLUMN `birthday` datetime(6) DEFAULT NULL,
                MODIFY COLUMN `deathday` datetime(6) DEFAULT NULL;
            ''')


class ParseCreateTableTests(SimpleTestCase):

    def test_large_normalish_table(self):
        sql = dedent("""\
        CREATE TABLE `example` (
          `id` int(11) NOT NULL AUTO_INCREMENT,
          `varchary` varchar(255) NOT NULL,
          `texty` longtext NOT NULL,
          `inty` int(11) DEFAULT NULL,
          `datetimey` datetime NOT NULL,
          PRIMARY KEY (`id`),
          KEY `example_abcdef` (`varchary`),
          KEY `example_123456` (`inty`)
          CONSTRAINT `example_constraint` FOREIGN KEY (`id`) REFERENCES `other` (`id`)
        ) ENGINE=InnoDB AUTO_INCREMENT=999999 DEFAULT CHARSET=utf8mb4 ROW_FORMAT=COMPRESSED
        """)  # noqa

        assert parse_create_table(sql) == {
            'id': 'int(11) NOT NULL AUTO_INCREMENT',
            'varchary': 'varchar(255) NOT NULL',
            'texty': 'longtext NOT NULL',
            'inty': 'int(11) DEFAULT NULL',
            'datetimey': 'datetime NOT NULL',
        }

    def test_single_column(self):
        sql = dedent("""\
        CREATE TABLE `example_table` (
          `the_data` longtext COLLATE utf8mb4_unicode_ci NOT NULL
        ) ENGINE=MyISAM DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_unicode_ci COMMENT='hi mum';
        """)  # noqa

        assert parse_create_table(sql) == {
            'the_data': 'longtext COLLATE utf8mb4_unicode_ci NOT NULL',
        }



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

from django.core.checks import Tags, Warning, register
from django.db import DEFAULT_DB_ALIAS, connections

from django_mysql.utils import collapse_spaces


def register_checks():
    register(Tags.compatibility)(check_variables)


def check_variables(app_configs, **kwargs):
    errors = []

    for alias, connection in mysql_connections():
        with connection.temporary_connection() as cursor:
            cursor.execute("""SELECT @@sql_mode,
                                     @@innodb_strict_mode,
                                     @@character_set_connection""")
            variables = cursor.fetchone()
            sql_mode, innodb_strict_mode, character_set_connection = variables

        modes = set(sql_mode.split(','))
        if not (modes & {'STRICT_TRANS_TABLES', 'STRICT_ALL_TABLES'}):
            errors.append(strict_mode_warning(alias))

        if not innodb_strict_mode:
            errors.append(innodb_strict_mode_warning(alias))

        if character_set_connection != 'utf8mb4':
            errors.append(utf8mb4_warning(alias))

    return errors


def strict_mode_warning(alias):
    message = "MySQL Strict Mode is not set for database connection '{}'"
    hint = collapse_spaces("""
        MySQL's Strict Mode fixes many data integrity problems in MySQL, such
        as data truncation upon insertion, by escalating warnings into errors.
        It is strongly recommended you activate it. See:
        https://django-mysql.readthedocs.io/en/latest/checks.html#django-mysql-w001-strict-mode
    """)
    return Warning(
        message.format(alias),
        hint=hint,
        id='django_mysql.W001',
    )


def innodb_strict_mode_warning(alias):
    message = "InnoDB Strict Mode is not set for database connection '{}'"
    hint = collapse_spaces("""
        InnoDB Strict Mode escalates several warnings around InnoDB-specific
        statements into errors. It's recommended you activate this, but it's
        not very likely to affect you if you don't. See:
        https://django-mysql.readthedocs.io/en/latest/checks.html#django-mysql-w002-innodb-strict-mode
    """)

    return Warning(
        message.format(alias),
        hint=hint,
        id='django_mysql.W002',
    )


def utf8mb4_warning(alias):
    message = "The character set is not utf8mb4 for database connection '{}'"
    hint = collapse_spaces("""
        The default 'utf8' character set does not include support for all
        Unicode characters. It's strongly recommended you move to use
        'utf8mb4'. See:
        https://django-mysql.readthedocs.io/en/latest/checks.html#django-mysql-w003-utf8mb4
    """)

    return Warning(
        message.format(alias),
        hint=hint,
        id='django_mysql.W003',
    )


def mysql_connections():
    conn_names = [DEFAULT_DB_ALIAS] + list(
        set(connections) - {DEFAULT_DB_ALIAS},
    )
    for alias in conn_names:
        connection = connections[alias]
        if not hasattr(connection, 'mysql_version'):
            continue  # pragma: no cover

        yield alias, connection



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import warnings

import django
from django.db import connection
from pytest_django.plugin import _blocking_manager


def pytest_report_header(config):
    dot_version = '.'.join(str(x) for x in django.VERSION)
    header = "Django version: " + dot_version

    with _blocking_manager.unblock():
        with connection._nodb_connection.cursor() as cursor:
            cursor.execute("SELECT VERSION()")
            version = cursor.fetchone()[0]
        header += "\nMySQL version: {}".format(version)

    return header


# MySQL 5.7 warns about some sql mode changes
warnings.filterwarnings('ignore', r'.*Changing sql mode.*')
warnings.filterwarnings(
    'ignore',
    r'.*sql modes should be used with strict mode.*',
)
# MySQL 5.7 turned 'explain' into 'explain extended' so it always warns the
# optimized query
warnings.filterwarnings('ignore', r'.*/\* select#\d+ \*/')
# MySQL 5.7 deprecated some query hints
warnings.filterwarnings('ignore', r".*'SQL_CACHE' is deprecated.*")
warnings.filterwarnings('ignore', r".*'SQL_NO_CACHE' is deprecated.*")



# -*- coding: utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

__author__ = 'Adam Johnson'
__email__ = 'me@adamj.eu'
__version__ = '2.5.0'


default_app_config = 'django_mysql.apps.MySQLConfig'



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

from django.db import models

from django_mysql.models.query import QuerySet


class Model(models.Model):
    class Meta(object):
        abstract = True

    objects = QuerySet.as_manager()



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import pytest
from django.core.management import CommandError, call_command
from django.db.utils import ConnectionHandler
from django.test import SimpleTestCase
from django.utils.six.moves import StringIO

try:
    from unittest import mock
except ImportError:
    import mock

# Can't use @override_settings to swap out DATABASES, instead just mock.patch
# a new ConnectionHandler into the command module
command_connections = 'django_mysql.management.commands.dbparams.connections'

sqlite = ConnectionHandler({
    'default': {'ENGINE': 'django.db.backends.sqlite3'},
})

full_db = ConnectionHandler({'default': {
    'ENGINE': 'django.db.backends.mysql',
    'NAME': 'mydatabase',
    'USER': 'ausername',
    'PASSWORD': 'apassword',
    'HOST': 'ahost.example.com',
    'PORT': '12345',
    'OPTIONS': {
        'read_default_file': '/tmp/defaults.cnf',
        'ssl': {'ca': '/tmp/mysql.cert'},
    },
}})

socket_db = ConnectionHandler({'default': {
    'ENGINE': 'django.db.backends.mysql',
    'HOST': '/etc/mydb.sock',
}})


class DBParamsTests(SimpleTestCase):

    def test_invalid_database(self):
        with pytest.raises(CommandError) as excinfo:
            call_command('dbparams', 'nonexistent', skip_checks=True)
        assert "does not exist" in str(excinfo.value)

    def test_invalid_both(self):
        with pytest.raises(CommandError):
            call_command('dbparams', dsn=True, mysql=True, skip_checks=True)

    @mock.patch(command_connections, sqlite)
    def test_invalid_not_mysql(self):
        with pytest.raises(CommandError) as excinfo:
            call_command('dbparams', skip_checks=True)
        assert "not a MySQL database connection" in str(excinfo.value)

    @mock.patch(command_connections, full_db)
    def test_mysql_full(self):
        out = StringIO()
        call_command('dbparams', stdout=out, skip_checks=True)
        output = out.getvalue()
        assert (
            output
            == (
                "--defaults-file=/tmp/defaults.cnf --user=ausername "
                + "--password=apassword --host=ahost.example.com --port=12345 "
                + "--ssl-ca=/tmp/mysql.cert mydatabase"
            )
        )

    @mock.patch(command_connections, socket_db)
    def test_mysql_socket(self):
        out = StringIO()
        call_command('dbparams', stdout=out, skip_checks=True)
        output = out.getvalue()
        assert output == "--socket=/etc/mydb.sock"

    @mock.patch(command_connections, full_db)
    def test_dsn_full(self):
        out = StringIO()
        err = StringIO()
        call_command('dbparams', 'default', dsn=True,
                     stdout=out, stderr=err, skip_checks=True)
        output = out.getvalue()
        assert (
            output
            == (
                "F=/tmp/defaults.cnf,u=ausername,p=apassword,"
                + "h=ahost.example.com,P=12345,D=mydatabase"
            )
        )

        errors = err.getvalue()
        assert "SSL params can't be" in errors

    @mock.patch(command_connections, socket_db)
    def test_dsn_socket(self):
        out = StringIO()
        err = StringIO()
        call_command('dbparams', dsn=True,
                     stdout=out, stderr=err, skip_checks=True)

        output = out.getvalue()
        assert output == 'S=/etc/mydb.sock'

        errors = err.getvalue()
        assert errors == ""



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import hashlib
from unittest import SkipTest

import pytest
from django.db import connection
from django.db.models import F, FloatField, Q, Value
from django.db.models.functions import Length, Lower, Upper
from django.test import TestCase
from django.utils import six

from django_mysql.models.functions import (
    CRC32, ELT, MD5, SHA1, SHA2, Abs, AsType, Ceiling, ColumnAdd, ColumnDelete,
    ColumnGet, ConcatWS, Field, Floor, Greatest, If, JSONArrayAppend,
    JSONExtract, JSONInsert, JSONKeys, JSONLength, JSONReplace, JSONSet,
    LastInsertId, Least, RegexpInstr, RegexpReplace, RegexpSubstr, Round, Sign,
    UpdateXML, XMLExtractValue,
)
from django_mysql.utils import connection_is_mariadb
from testapp.models import Alphabet, Author, DynamicModel, JSONModel
from testapp.test_dynamicfield import DynColTestCase
from testapp.test_jsonfield import JSONFieldTestCase


class ComparisonFunctionTests(TestCase):

    def test_greatest(self):
        Alphabet.objects.create(d='A', e='B')
        ab = Alphabet.objects.annotate(best=Greatest('d', 'e')).first()
        assert ab.best == 'B'

    def test_greatest_takes_no_kwargs(self):
        with pytest.raises(TypeError):
            Greatest('a', something='wrong')

    def test_least(self):
        Alphabet.objects.create(a=1, b=2, c=-1)
        ab = Alphabet.objects.annotate(worst=Least('a', 'b', 'c')).first()
        assert ab.worst == -1


class ControlFlowFunctionTests(TestCase):

    def test_if_basic(self):
        Alphabet.objects.create(d='String')
        Alphabet.objects.create(d='')
        Alphabet.objects.create(d='String')
        Alphabet.objects.create(d='')

        results = list(Alphabet.objects.annotate(
            has_d=If(Length('d'), Value(True), Value(False)),
        ).order_by('id').values_list('has_d', flat=True))
        assert results == [True, False, True, False]

    def test_if_with_Q(self):
        Alphabet.objects.create(a=12, b=17)
        Alphabet.objects.create(a=13, b=17)
        Alphabet.objects.create(a=14, b=17)

        result = list(Alphabet.objects.annotate(
            conditional=If(Q(a__lte=13), 'a', 'b'),
        ).order_by('id').values_list('conditional', flat=True))
        assert result == [12, 13, 17]

    def test_if_with_string_values(self):
        Alphabet.objects.create(a=1, d='Lentils')
        Alphabet.objects.create(a=2, d='Cabbage')
        Alphabet.objects.create(a=3, d='Rice')

        result = list(Alphabet.objects.annotate(
            conditional=If(Q(a=2), Upper('d'), Lower('d')),
        ).order_by('id').values_list('conditional', flat=True))
        assert result == ['lentils', 'CABBAGE', 'rice']

    def test_if_field_lookups_work(self):
        Alphabet.objects.create(a=1, d='Lentils')
        Alphabet.objects.create(a=2, d='Cabbage')
        Alphabet.objects.create(a=3, d='Rice')

        result = list(Alphabet.objects.annotate(
            conditional=If(Q(a__gte=2), Upper('d'), Value('')),
        ).filter(
            conditional__startswith='C',
        ).order_by('id').values_list('conditional', flat=True))
        assert result == ['CABBAGE']

    def test_if_false_default_None(self):
        Alphabet.objects.create(a=1)

        result = list(Alphabet.objects.annotate(
            conditional=If(Q(a=2), Value(1)),
        ).filter(
            conditional__isnull=True,
        ).values_list('conditional', flat=True))
        assert result == [None]


class NumericFunctionTests(TestCase):

    def test_abs(self):
        Alphabet.objects.create(a=-2)
        ab = Alphabet.objects.annotate(aaa=Abs('a')).first()
        assert ab.aaa == 2

    def test_ceiling(self):
        Alphabet.objects.create(g=0.5)
        ab = Alphabet.objects.annotate(gceil=Ceiling('g')).first()
        assert ab.gceil == 1

    def test_crc32(self):
        Alphabet.objects.create(d='AAAAAA')
        ab = Alphabet.objects.annotate(crc=CRC32('d')).first()
        # Precalculated this in MySQL prompt. Python's binascii.crc32 doesn't
        # match - maybe sign issues?
        assert ab.crc == 2854018686

    def test_crc32_only_takes_one_arg_no_kwargs(self):
        with pytest.raises(TypeError):
            CRC32('d', 'c')

        with pytest.raises(TypeError):
            CRC32('d', something='wrong')

    def test_floor(self):
        Alphabet.objects.create(g=1.5)
        ab = Alphabet.objects.annotate(gfloor=Floor('g')).first()
        assert ab.gfloor == 1

    def test_round(self):
        Alphabet.objects.create(g=24.459)
        ab = Alphabet.objects.annotate(ground=Round('g')).get()
        assert ab.ground == 24

    def test_round_up(self):
        Alphabet.objects.create(g=27.859)
        ab = Alphabet.objects.annotate(ground=Round('g')).get()
        assert ab.ground == 28

    def test_round_places(self):
        Alphabet.objects.create(a=81731)
        ab = Alphabet.objects.annotate(around=Round('a', -2)).get()
        assert ab.around == 81700

    def test_sign(self):
        Alphabet.objects.create(a=123, b=0, c=-999)
        ab = Alphabet.objects.annotate(
            asign=Sign('a'),
            bsign=Sign('b'),
            csign=Sign('c'),
        ).first()
        assert ab.asign == 1
        assert ab.bsign == 0
        assert ab.csign == -1


class StringFunctionTests(TestCase):

    def test_concat_ws(self):
        Alphabet.objects.create(d='AAA', e='BBB')
        ab = Alphabet.objects.annotate(de=ConcatWS('d', 'e')).first()
        assert ab.de == 'AAA,BBB'

    def test_concat_ws_integers(self):
        Alphabet.objects.create(a=1, b=2)
        ab = Alphabet.objects.annotate(ab=ConcatWS('a', 'b')).first()
        assert ab.ab == '1,2'

    def test_concat_ws_skips_nulls(self):
        Alphabet.objects.create(d='AAA', e=None, f=2)
        ab = Alphabet.objects.annotate(de=ConcatWS('d', 'e', 'f')).first()
        assert ab.de == 'AAA,2'

    def test_concat_ws_separator(self):
        Alphabet.objects.create(d='AAA', e='BBB')
        ab = (
            Alphabet.objects.annotate(de=ConcatWS('d', 'e', separator=':'))
                            .first()
        )
        assert ab.de == 'AAA:BBB'

    def test_concat_ws_separator_null_returns_none(self):
        Alphabet.objects.create(a=1, b=2)
        concat = ConcatWS('a', 'b', separator=None)
        ab = Alphabet.objects.annotate(ab=concat).first()
        assert ab.ab is None

    def test_concat_ws_separator_field(self):
        Alphabet.objects.create(a=1, d='AAA', e='BBB')
        concat = ConcatWS('d', 'e', separator=F('a'))
        ab = Alphabet.objects.annotate(de=concat).first()
        assert ab.de == 'AAA1BBB'

    def test_concat_ws_bad_arg(self):
        with pytest.raises(ValueError) as excinfo:
            ConcatWS('a', 'b', separataaa=',')
        assert ('Invalid keyword arguments for ConcatWS: separataaa' in
                str(excinfo.value))

    def test_concat_ws_too_few_fields(self):
        with pytest.raises(ValueError) as excinfo:
            ConcatWS('a')
        assert ('ConcatWS must take at least two expressions' in
                str(excinfo.value))

    def test_concat_ws_then_lookups_from_textfield(self):
        Alphabet.objects.create(d='AAA', e='BBB')
        Alphabet.objects.create(d='AAA', e='CCC')
        ab = (
            Alphabet.objects.annotate(de=ConcatWS('d', 'e', separator=':'))
                            .filter(de__endswith=':BBB')
                            .first()
        )
        assert ab.de == 'AAA:BBB'

    def test_elt_simple(self):
        Alphabet.objects.create(a=2)
        ab = Alphabet.objects.annotate(elt=ELT('a', ['apple', 'orange'])).get()
        assert ab.elt == 'orange'
        ab = Alphabet.objects.annotate(elt=ELT('a', ['apple'])).get()
        assert ab.elt is None

    def test_field_simple(self):
        Alphabet.objects.create(d='a')
        ab = Alphabet.objects.annotate(dp=Field('d', ['a', 'b'])).first()
        assert ab.dp == 1
        ab = Alphabet.objects.annotate(dp=Field('d', ['b', 'a'])).first()
        assert ab.dp == 2
        ab = Alphabet.objects.annotate(dp=Field('d', ['c', 'd'])).first()
        assert ab.dp == 0

    def test_order_by(self):
        Alphabet.objects.create(a=1, d='AAA')
        Alphabet.objects.create(a=2, d='CCC')
        Alphabet.objects.create(a=4, d='BBB')
        Alphabet.objects.create(a=3, d='BBB')
        avalues = list(
            Alphabet.objects.order_by(Field('d', ['AAA', 'BBB']), 'a')
                            .values_list('a', flat=True),
        )
        assert avalues == [2, 1, 3, 4]


class XMLFunctionTests(TestCase):

    def test_updatexml_simple(self):
        Alphabet.objects.create(d='<value>123</value>')
        Alphabet.objects.update(
            d=UpdateXML('d', '/value', '<value>456</value>'),
        )
        d = Alphabet.objects.get().d
        assert d == '<value>456</value>'

    def test_updatexml_annotate(self):
        Alphabet.objects.create(d='<value>123</value>')
        d2 = Alphabet.objects.annotate(
            d2=UpdateXML('d', '/value', '<value>456</value>'),
        ).get().d2
        assert d2 == '<value>456</value>'

    def test_xmlextractvalue_simple(self):
        Alphabet.objects.create(d='<some><xml /></some>')
        Alphabet.objects.create(d='<someother><xml /></someother>')
        Alphabet.objects.create(d='<some></some><some></some>')
        evs = list(
            Alphabet.objects.annotate(ev=XMLExtractValue('d', 'count(/some)'))
                            .values_list('ev', flat=True),
        )
        assert evs == ['1', '0', '2']

    def test_xmlextractvalue_invalid_xml(self):
        Alphabet.objects.create(d='{"this": "isNotXML"}')
        ab = (
            Alphabet.objects.annotate(ev=XMLExtractValue('d', '/some'))
                            .first()
        )
        assert ab.ev == ''


class EncryptionFunctionTests(TestCase):

    def test_md5_string(self):
        string = 'A string'
        Alphabet.objects.create(d=string)
        pymd5 = hashlib.md5(string.encode('ascii')).hexdigest()
        ab = Alphabet.objects.annotate(md5=MD5('d')).first()
        assert ab.md5 == pymd5

    def test_sha1_string(self):
        string = 'A string'
        Alphabet.objects.create(d=string)
        pysha1 = hashlib.sha1(string.encode('ascii')).hexdigest()
        ab = Alphabet.objects.annotate(sha=SHA1('d')).first()
        assert ab.sha == pysha1

    def test_sha2_string(self):
        string = 'A string'
        Alphabet.objects.create(d=string)

        for hash_len in (224, 256, 384, 512):
            sha_func = getattr(hashlib, 'sha{}'.format(hash_len))
            pysha = sha_func(string.encode('ascii')).hexdigest()
            ab = Alphabet.objects.annotate(sha=SHA2('d', hash_len)).first()
            assert ab.sha == pysha

    def test_sha2_string_hash_len_default(self):
        string = 'A string'
        Alphabet.objects.create(d=string)
        pysha512 = hashlib.sha512(string.encode('ascii')).hexdigest()
        ab = Alphabet.objects.annotate(sha=SHA2('d')).first()
        assert ab.sha == pysha512

    def test_sha2_bad_hash_len(self):
        with pytest.raises(ValueError):
            SHA2('a', 123)


class InformationFunctionTests(TestCase):

    multi_db = True

    def test_last_insert_id(self):
        Alphabet.objects.create(a=7891)
        Alphabet.objects.update(a=LastInsertId('a') + 1)
        lid = LastInsertId.get()
        assert lid == 7891

    def test_last_insert_id_other_db_connection(self):
        Alphabet.objects.using('other').create(a=9191)
        Alphabet.objects.using('other').update(a=LastInsertId('a') + 9)
        lid = LastInsertId.get(using='other')
        assert lid == 9191

    def test_last_insert_id_in_query(self):
        ab1 = Alphabet.objects.create(a=3719, b=717612)
        ab2 = Alphabet.objects.create(a=1838, b=12636)

        # Delete but store value of b and re-assign it to first second Alphabet
        Alphabet.objects.filter(id=ab1.id, b=LastInsertId('b')).delete()
        Alphabet.objects.filter(id=ab2.id).update(b=LastInsertId())

        ab = Alphabet.objects.get()
        assert ab.b == 717612


class JSONFunctionTests(JSONFieldTestCase):

    def setUp(self):
        super(JSONFunctionTests, self).setUp()
        self.obj = JSONModel.objects.create(attrs={
            'int': 88,
            'flote': 1.5,
            'sub': {
                'document': 'store',
            },
            'arr': ['dee', 'arr', 'arr'],
        })

    def test_json_extract_kwarg_bad(self):
        with pytest.raises(TypeError) as excinfo:
            JSONExtract('foo', 'bar', foo=1)
        assert 'Only supported keyword' in six.text_type(excinfo.value)

    def test_json_extract_output_field_too_many_paths(self):
        with pytest.raises(TypeError) as excinfo:
            JSONExtract('foo', 'bar', 'baz', output_field=1)
        assert (
            "output_field won't work with more than one path" in
            six.text_type(excinfo.value)
        )

    def test_json_extract_flote(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONExtract('attrs', '$.flote'),
            ).values_list('x', flat=True),
        )
        assert results == [1.5]
        assert isinstance(results[0], float)

    def test_json_extract_flote_as_float(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONExtract('attrs', '$.flote', output_field=FloatField()),
            ).filter(
                x__gt=0.1,
            ).values_list('x', flat=True),
        )
        assert results == [1.5]
        assert isinstance(results[0], float)

    def test_json_extract_multiple(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONExtract('attrs', '$.int', '$.flote'),
            ).values_list('x', flat=True),
        )
        assert results == [[88, 1.5]]
        assert isinstance(results[0][0], six.integer_types)
        assert isinstance(results[0][1], float)

    def test_json_extract_filter(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONExtract('attrs', '$.sub'),
            ).filter(
                x={'document': 'store'},
            ),
        )
        assert results == [self.obj]

    def test_json_keys(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONKeys('attrs'),
            ).values_list('x', flat=True),
        )
        assert set(results[0]) == set(self.obj.attrs.keys())

    def test_json_keys_path(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONKeys('attrs', '$.sub'),
            ).values_list('x', flat=True),
        )
        assert set(results[0]) == set(self.obj.attrs['sub'].keys())

    def test_json_length(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONLength('attrs'),
            ).values_list('x', flat=True),
        )
        assert results == [len(self.obj.attrs)]

    def test_json_length_path(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONLength('attrs', '$.sub'),
            ).values_list('x', flat=True),
        )
        assert results == [len(self.obj.attrs['sub'])]

    def test_json_length_type(self):
        results = list(
            JSONModel.objects.annotate(
                x=JSONLength('attrs'),
            ).filter(x__range=[3, 5]),
        )
        assert results == [self.obj]

    def test_json_insert(self):
        self.obj.attrs = JSONInsert('attrs', {'$.int': 99, '$.int2': 102})
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['int'] == 88
        assert self.obj.attrs['int2'] == 102

    def test_json_insert_dict(self):
        self.obj.attrs = JSONInsert(
            'attrs',
            {'$.sub': {'paper': 'drop'}, '$.sub2': {'int': 42, 'foo': 'bar'}},
        )
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['sub'] == {'document': 'store'}
        assert self.obj.attrs['sub2']['int'] == 42
        assert self.obj.attrs['sub2']['foo'] == 'bar'

    def test_json_insert_array(self):
        self.obj.attrs = JSONInsert(
            'attrs',
            {'$.arr': [1, 'two', 3], '$.arr2': ['one', 2]},
        )
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['arr'] == ['dee', 'arr', 'arr']
        assert self.obj.attrs['arr2'][0] == 'one'
        assert self.obj.attrs['arr2'][1] == 2

    def test_json_insert_empty_data(self):
        with pytest.raises(ValueError) as excinfo:
            JSONInsert('attrs', {})
        assert '"data" cannot be empty' in str(excinfo.value)

    def test_json_replace_pairs(self):
        self.obj.attrs = JSONReplace('attrs', {'$.int': 101, '$.int2': 102})
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['int'] == 101
        assert 'int2' not in self.obj.attrs

    def test_json_replace_dict(self):
        self.obj.attrs = JSONReplace(
            'attrs',
            {'$.sub': {'paper': 'drop'}, '$.sub2': {'int': 42, 'foo': 'bar'}},
        )
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['sub'] == {'paper': 'drop'}
        assert 'sub2' not in self.obj.attrs

    def test_json_replace_array(self):
        self.obj.attrs = JSONReplace(
            'attrs',
            {'$.arr': [1, 'two', 3], '$.arr2': ['one', 2]},
        )
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['arr'] == [1, 'two', 3]
        assert 'arr2' not in self.obj.attrs

    def test_json_replace_empty_data(self):
        with pytest.raises(ValueError) as excinfo:
            JSONReplace('attrs', {})
        assert '"data" cannot be empty' in str(excinfo.value)

    def test_json_set_pairs(self):
        self.obj.attrs = JSONSet('attrs', {'$.int': 101, '$.int2': 102})
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['int'] == 101
        assert self.obj.attrs['int2'] == 102

    def test_json_set_dict(self):
        self.obj.attrs = JSONSet(
            'attrs',
            {'$.sub': {'paper': 'drop'}, '$.sub2': {'int': 42, 'foo': 'bar'}},
        )
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['sub'] == {'paper': 'drop'}
        assert self.obj.attrs['sub2']['int'] == 42
        assert self.obj.attrs['sub2']['foo'] == 'bar'

    def test_json_set_array(self):
        self.obj.attrs = JSONSet(
            'attrs',
            {'$.arr': [1, 'two', 3], '$.arr2': ['one', 2]},
        )
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['arr'] == [1, 'two', 3]
        assert self.obj.attrs['arr2'][0] == 'one'
        assert self.obj.attrs['arr2'][1] == 2

    def test_json_set_complex_data(self):
        data = {
            'list': ['one', {'int': 2}, None],
            'dict': {
                'sub_list': [1, ['two', 3]],
                'sub_dict': {'paper': 'drop'},
                'sub_null': None,
            },
            'empty': [],
            'null': None,
            'value': 11,
        }
        self.obj.attrs = JSONSet('attrs', {'$.data': data})
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['data'] == data

    def test_json_set_empty_data(self):
        with pytest.raises(ValueError) as excinfo:
            JSONSet('attrs', {})
        assert '"data" cannot be empty' in str(excinfo.value)

    def test_json_array_append(self):
        self.obj.attrs = JSONArrayAppend(
            'attrs',
            {'$.arr': 'max', '$.arr[0]': 1.1, '$.sub.document': 3},
        )
        self.obj.save()
        self.obj.refresh_from_db()
        assert self.obj.attrs['arr'] == [['dee', 1.1], 'arr', 'arr', 'max']
        assert self.obj.attrs['sub']['document'] == ['store', 3]


class RegexpFunctionTests(TestCase):

    def setUp(self):
        super(RegexpFunctionTests, self).setUp()
        have_regex_functions = (
            connection_is_mariadb(connection)
            and connection.mysql_version >= (10, 0, 5)
        )
        if not have_regex_functions:
            raise SkipTest("MariaDB 10.0.5+ is required")

    def test_regex_instr(self):
        Alphabet.objects.create(d="ABC")
        Alphabet.objects.create(d="ABBC")
        ab = (
            Alphabet.objects.annotate(d_double_pos=RegexpInstr('d', r'[b]{2}'))
                            .filter(d_double_pos__gt=0)
                            .get()
        )
        assert ab.d == "ABBC"
        assert ab.d_double_pos == 2

    def test_regex_instr_update(self):
        Alphabet.objects.create(d="A string to search")
        Alphabet.objects.create(d="Something to query")
        Alphabet.objects.create(d="Please do inspect me")
        Alphabet.objects.update(
            a=RegexpInstr('d', r'search|query|inspect') - 1,
        )
        all_abs = Alphabet.objects.all().order_by('id')
        assert all_abs[0].a == 12
        assert all_abs[1].a == 13
        assert all_abs[2].a == 10

    def test_regex_replace_update(self):
        Alphabet.objects.create(d="I'm feeling sad")
        n = Alphabet.objects.update(d=RegexpReplace('d', r'\bsad\b', 'happy'))
        assert n == 1
        ab = Alphabet.objects.get()
        assert ab.d == "I'm feeling happy"

    def test_regex_replace_update_groups(self):
        Alphabet.objects.create(d="A,,List,with,,empty,,,,strings, ")
        n = Alphabet.objects.update(d=RegexpReplace('d', r',{2,}', ','))
        assert n == 1
        ab = Alphabet.objects.get()
        assert ab.d == "A,List,with,empty,strings, "

    def test_regex_replace_filter_backref(self):
        Author.objects.create(name="Charles Dickens")
        Author.objects.create(name="Roald Dahl")
        qs = (
            Author.objects.annotate(
                surname_first=RegexpReplace('name', r'^(.*) (.*)$', r'\2, \1'),
            ).order_by('surname_first')
        )
        assert qs[0].name == "Roald Dahl"
        assert qs[1].name == "Charles Dickens"

    def test_regex_substr(self):
        Alphabet.objects.create(d='The <tag> is there')
        qs = (
            Alphabet.objects.annotate(d_tag=RegexpSubstr('d', r'<[^>]+>'))
                            .filter(d_tag__gt='')
        )
        ab = qs.get()
        assert ab.d_tag == '<tag>'

    def test_regex_substr_field(self):
        Alphabet.objects.create(d='This is the string', e=r'\bis\b')
        qs = (
            Alphabet.objects.annotate(substr=RegexpSubstr('d', F('e')))
                            .filter(substr__gt='')
                            .values_list('substr', flat=True)
        )
        substr = qs[0]
        assert substr == 'is'

    def test_regex_substr_filter(self):
        Author.objects.create(name="Euripides")
        Author.objects.create(name="Frank Miller")
        Author.objects.create(name="Sophocles")
        qs = list(
            Author.objects.annotate(
                name_has_space=Length(RegexpSubstr('name', r'\s')),
            ).filter(name_has_space=0)
            .order_by('id').values_list('name', flat=True),
        )
        assert qs == ['Euripides', 'Sophocles']


class DynamicColumnsFunctionTests(DynColTestCase):

    def setUp(self):
        super(DynamicColumnsFunctionTests, self).setUp()
        DynamicModel.objects.create(attrs={
            'flote': 1.0,
            'sub': {
                'document': 'store',
            },
        })

    def test_get_float(self):
        results = list(
            DynamicModel.objects.annotate(
                x=ColumnGet('attrs', 'flote', 'DOUBLE'),
            ).values_list('x', flat=True),
        )
        assert results == [1.0]
        assert isinstance(results[0], float)

    def test_get_int(self):
        results = list(
            DynamicModel.objects.annotate(
                x=ColumnGet('attrs', 'flote', 'INTEGER'),
            ).values_list('x', flat=True),
        )
        assert results == [1]
        assert isinstance(results[0], int)

    def test_get_null(self):
        results = list(
            DynamicModel.objects.annotate(
                x=ColumnGet('attrs', 'nonexistent', 'INTEGER'),
            ).values_list('x', flat=True),
        )
        assert results == [None]

    def test_get_nested(self):
        results = list(
            DynamicModel.objects.annotate(
                x=ColumnGet(
                    ColumnGet('attrs', 'sub', 'BINARY'),
                    'document',
                    'CHAR',
                ),
            ).values_list('x', flat=True),
        )
        assert results == ['store']

    def test_get_invalid_data_type(self):
        with pytest.raises(ValueError) as excinfo:
            ColumnGet('bla', 'key', 'INTGRRR')
        assert "Invalid data_type 'INTGRRR'" in str(excinfo.value)

    def test_add(self):
        results = list(
            DynamicModel.objects.annotate(
                attrs2=ColumnAdd('attrs', {'another': 'key'}),
            ).values_list('attrs2', flat=True),
        )
        assert results == [{
            'flote': 1.0,
            'sub': {'document': 'store'},
            'another': 'key',
        }]

    def test_add_nested(self):
        with pytest.raises(ValueError) as excinfo:
            list(
                DynamicModel.objects.annotate(
                    attrs2=ColumnAdd('attrs', {'sub2': {'document': 'store'}}),
                ).values_list('attrs2', flat=True),
            )
        assert "nested values is not supported" in str(excinfo.value)

    def test_add_update(self):
        DynamicModel.objects.update(attrs=ColumnAdd('attrs', {'over': 9000}))
        m = DynamicModel.objects.get()
        assert m.attrs == {
            'flote': 1.0,
            'sub': {'document': 'store'},
            'over': 9000,
        }

    def test_as_type_instantiation(self):
        with pytest.raises(ValueError) as excinfo:
            AsType(1, 'PANTS')
        assert "Invalid data_type 'PANTS'" in str(excinfo.value)

    def test_add_update_typed(self):
        DynamicModel.objects.update(
            attrs=ColumnAdd('attrs', {'over': AsType(9000, 'DOUBLE')}),
        )
        m = DynamicModel.objects.get()
        assert isinstance(m.attrs['over'], float)
        assert m.attrs['over'] == 9000.0

    def test_delete(self):
        DynamicModel.objects.update(attrs=ColumnDelete('attrs', 'sub'))
        m = DynamicModel.objects.get()
        assert m.attrs == {'flote': 1.0}

    def test_delete_subfunc(self):
        say_sub = ConcatWS(Value('s'), Value('ub'), separator='')
        DynamicModel.objects.update(attrs=ColumnDelete('attrs', say_sub))
        m = DynamicModel.objects.get()
        assert m.attrs == {'flote': 1.0}



# -*- encoding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

from django.db.models import CharField
from django.utils import six
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _


class EnumField(CharField):
    description = _("Enumeration")

    def __init__(self, *args, **kwargs):
        if 'choices' not in kwargs or len(kwargs['choices']) == 0:
            raise ValueError(
                '"choices" argument must be be a non-empty list',
            )

        choices = []
        for choice in kwargs['choices']:
            if isinstance(choice, tuple):
                choices.append(choice)
            elif isinstance(choice, six.string_types):
                choices.append((choice, choice))
            else:
                raise TypeError(
                    'Invalid choice "{choice}". '
                    'Expected string or tuple as elements in choices'.format(
                        choice=choice,
                    ),
                )

        kwargs['choices'] = choices

        if 'max_length' in kwargs:
            raise TypeError('"max_length" is not a valid argument')
        # Massive to avoid problems with validation - let MySQL handle the
        # maximum string length
        kwargs['max_length'] = int(2 ** 32)

        super(EnumField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(EnumField, self).deconstruct()

        bad_paths = (
            'django_mysql.models.fields.enum.EnumField',
            'django_mysql.models.fields.EnumField',
        )
        if path in bad_paths:
            path = 'django_mysql.models.EnumField'

        kwargs['choices'] = self.choices
        del kwargs['max_length']

        return name, path, args, kwargs

    def db_type(self, connection):
        connection.ensure_connection()
        values = [
            connection.connection.escape_string(c)
            for c, _ in self.flatchoices
        ]
        # Use force_text because MySQLdb escape_string() returns bytes, but
        # pymysql returns str
        return 'enum(%s)' % ','.join(
            "'%s'" % force_text(v)
            for v in values
        )



# -*- coding: utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

from django.db import migrations, models

from django_mysql.models import EnumField


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='EnumDefaultModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False,
                 auto_created=True, primary_key=True)),
                ('field', EnumField(choices=[
                    ('lion', 'Lion'), ('tiger', 'Tiger'), 'bear',
                ])),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import pytest
from django.db import connection
from django.db.models import F
from django.test import TestCase

from django_mysql.models.handler import Handler
from django_mysql.utils import index_name
from testapp.models import (
    Author, AuthorHugeName, AuthorMultiIndex, NameAuthor, VanillaAuthor,
)


def get_index_names(model):
    # There's no easy way of getting index names in django so pull them from
    # INFORMATION_SCHEMA
    with connection.cursor() as cursor:
        cursor.execute(
            """SELECT DISTINCT INDEX_NAME
            FROM INFORMATION_SCHEMA.STATISTICS
            WHERE TABLE_SCHEMA = %s AND
                  TABLE_NAME = %s""",
            (connection.settings_dict['NAME'], model._meta.db_table),
        )

        index_names = [x[0] for x in cursor.fetchall()]
    return index_names


class HandlerCreationTests(TestCase):

    def test_bad_creation_joins_not_allowed(self):
        qs = Author.objects.filter(tutor__name='A')
        with pytest.raises(ValueError):
            qs.handler()

    def test_bad_creation_limit_not_allowed(self):
        qs = Author.objects.all()[:100]
        with pytest.raises(ValueError):
            qs.handler()

    def test_bad_creation_ordering_not_allowed(self):
        qs = Author.objects.order_by('name')
        with pytest.raises(ValueError):
            qs.handler()

    def test_can_open_close_with_huge_table_name(self):
        with AuthorHugeName.objects.handler():
            pass

    def test_cannot_open_twice(self):
        handler = Author.objects.handler()
        with handler:
            with pytest.raises(ValueError):
                with handler:
                    pass

    def test_cannot_close_unopened(self):
        handler = Author.objects.handler()
        with pytest.raises(ValueError):
            handler.__exit__(None, None, None)


class BaseAuthorTestCase(TestCase):

    def setUp(self):
        self.jk = Author.objects.create(name='JK Rowling')
        self.grisham = Author.objects.create(name='John Grisham')


class HandlerReadTests(BaseAuthorTestCase):

    def test_bad_read_unopened(self):
        handler = Author.objects.all().handler()
        with pytest.raises(RuntimeError):
            handler.read()

    def test_bad_read_mode(self):
        with Author.objects.handler() as handler:
            with pytest.raises(ValueError):
                handler.read(mode='non-existent')

    def test_read_does_single_by_default(self):
        with Author.objects.handler() as handler:
            handler_all = list(handler.read())
        assert handler_all == [self.jk]

    def test_read_limit_first(self):
        with Author.objects.handler() as handler:
            handler_first = handler.read(limit=1)[0]
        qs_first = Author.objects.earliest('id')

        assert handler_first == qs_first

    def test_read_limit_last(self):
        with Author.objects.handler() as handler:
            handler_last = handler.read(mode='last', limit=1)[0]
        qs_last = Author.objects.latest('id')

        assert handler_last == qs_last

    def test_read_limit_all(self):
        with Author.objects.handler() as handler:
            handler_all = list(handler.read(limit=2))
        qs_all = list(Author.objects.all())

        assert handler_all == qs_all

    def test_read_index_primary(self):
        with Author.objects.handler() as handler:
            handler_all = list(handler.read(index='PRIMARY', limit=2))
        qs_all = list(Author.objects.order_by('id'))

        assert handler_all == qs_all

    def test_read_index_different(self):
        index_name = [name for name in get_index_names(Author)
                      if name != "PRIMARY"][0]

        with Author.objects.handler() as handler:
            handler_all = list(handler.read(index=index_name, limit=2))
        qs_all = list(Author.objects.order_by('name').all())

        assert handler_all == qs_all

    def test_read_where_filter_read(self):
        qs = Author.objects.filter(name__startswith='John')

        with qs.handler() as handler:
            handler_all = list(handler.read())
        qs_all = list(qs)

        assert handler_all == qs_all

    def test_read_where_filter_f_expression(self):
        qs = Author.objects.filter(name=F('name'))

        with qs.handler() as handler:
            handler_all = list(handler.read(limit=100))

        assert len(handler_all) == 2

    def test_read_where_exclude(self):
        qs = Author.objects.filter(name__contains='JK')

        with qs.handler() as handler:
            handler_all = list(handler.read())
        qs_all = list(qs)

        assert handler_all == qs_all

    def test_read_where_filter_params_not_injected_or_modified(self):
        table_col = "`looks_like`.`table_column`"
        author = Author.objects.create(name=table_col)
        qs = Author.objects.filter(name=table_col)

        with qs.handler() as handler:
            handler_first = handler.read()[0]

        assert handler_first == author

    def test_read_where_passed_in(self):
        qs = Author.objects.filter(name__startswith='John')

        with Author.objects.handler() as handler:
            handler_author = handler.read(where=qs)[0]

        assert handler_author == qs[0]

    def test_read_where_passed_in_overrides_completely(self):
        qs = Author.objects.filter(name='JK Rowling')
        qs2 = Author.objects.filter(name='John Grisham')

        with qs.handler() as handler:
            handler_default = handler.read()[0]
            handler_where = handler.read(where=qs2)[0]

        assert handler_default == qs[0]
        assert handler_where == qs2[0]

    def test_read_bad_where_passed_in(self):
        with Author.objects.handler() as handler:
            with pytest.raises(ValueError):
                handler.read(where=Author.objects.filter(tutor__name='A'))

    def test_read_index_value_and_mode_invalid(self):
        with Author.objects.handler() as handler:
            with pytest.raises(ValueError):
                handler.read(value=1, mode='first')

    def test_read_index_equal(self):
        with Author.objects.handler() as handler:
            handler_result = handler.read(value=self.jk.id)[0]
        assert handler_result == self.jk

    def test_read_index_equal_exact(self):
        with Author.objects.handler() as handler:
            handler_result = handler.read(value__exact=self.jk.id)[0]
        assert handler_result == self.jk

    def test_read_index_less_than(self):
        with Author.objects.handler() as handler:
            handler_result = handler.read(value__lt=self.jk.id + 1)[0]
        assert handler_result == self.jk

    def test_read_index_less_than_equal(self):
        with Author.objects.handler() as handler:
            handler_result = handler.read(value__lte=self.jk.id)[0]
        assert handler_result == self.jk

    def test_read_index_greater_than_equal(self):
        with Author.objects.handler() as handler:
            handler_result = handler.read(value__gte=self.jk.id)[0]
        assert handler_result == self.jk

    def test_read_index_greater_than(self):
        with Author.objects.handler() as handler:
            handler_result = handler.read(value__gt=self.jk.id)[0]
        assert handler_result == self.grisham

    def test_read_index_too_many_filters(self):
        with Author.objects.handler() as handler:
            with pytest.raises(ValueError):
                handler.read(value__lte=1, value__gte=1)

    def test_read_index_bad_operator(self):
        with Author.objects.handler() as handler:
            with pytest.raises(ValueError):
                handler.read(value__flange=1)

    def test_read_bad_argument(self):
        with Author.objects.handler() as handler:
            with pytest.raises(ValueError):
                handler.read(pk=1)

    def test_read_bad_argument_underscores(self):
        with Author.objects.handler() as handler:
            with pytest.raises(ValueError):
                handler.read(value_exact=1)


class HandlerIterTests(BaseAuthorTestCase):

    def test_iter_all(self):
        with Author.objects.handler() as handler:
            seen_ids = [author.id for author in handler.iter()]

        assert seen_ids == [self.jk.id, self.grisham.id]

    def test_iter_chunk_size_1(self):
        with Author.objects.handler() as handler:
            seen_ids = [author.id for author in handler.iter(chunk_size=1)]

        assert seen_ids == [self.jk.id, self.grisham.id]

    def test_iter_reverse(self):
        with Author.objects.handler() as handler:
            seen_ids = [author.id for author in handler.iter(reverse=True)]

        assert seen_ids == [self.grisham.id, self.jk.id]

    def test_iter_reverse_chunk_size_1(self):
        with Author.objects.handler() as handler:
            seen_ids = [author.id for author in
                        handler.iter(chunk_size=1, reverse=True)]

        assert seen_ids == [self.grisham.id, self.jk.id]

    def test_bad_iter_unopened(self):
        handler = Author.objects.all().handler()
        with pytest.raises(RuntimeError):
            sum(1 for x in handler.iter())

    def test_iter_where_preset(self):
        where_qs = Author.objects.filter(name__startswith='John')

        with where_qs.handler() as handler:
            seen_ids = [author.id for author in handler.iter()]

        assert seen_ids == [self.grisham.id]

    def test_iter_where_passed_in(self):
        where_qs = Author.objects.filter(name__startswith='John')

        with Author.objects.handler() as handler:
            seen_ids = [author.id for author in handler.iter(where=where_qs)]

        assert seen_ids == [self.grisham.id]

    def test_iter_loses_its_place(self):
        portis = Author.objects.create(name='Charles Portis')

        with Author.objects.handler() as handler:
            iterator = handler.iter(chunk_size=1)
            assert next(iterator) == self.jk
            assert next(iterator) == self.grisham

            iterator2 = handler.iter(chunk_size=1, reverse=True)
            assert next(iterator2) == portis

            # This SHOULD be portis, but it thinks it's exhausted already
            # because iterator2 reset the iteration in reverse, and now for the
            # NEXT page after the LAST
            with pytest.raises(StopIteration):
                assert next(iterator) == portis


class HandlerMultipartIndexTests(TestCase):

    def setUp(self):
        super(HandlerMultipartIndexTests, self).setUp()
        self.smith1 = AuthorMultiIndex.objects.create(name='John Smith',
                                                      country='Scotland')
        self.smith2 = AuthorMultiIndex.objects.create(name='John Smith',
                                                      country='England')
        self.index_name = index_name(AuthorMultiIndex, 'name', 'country')

    def test_read_all(self):
        with AuthorMultiIndex.objects.handler() as handler:
            handler_all = list(handler.read(index=self.index_name, limit=2))
        qs_all = list(AuthorMultiIndex.objects.order_by('name', 'country'))
        assert handler_all == qs_all

    def test_read_index_multipart(self):
        with AuthorMultiIndex.objects.handler() as handler:
            value = ('John Smith', 'England')
            result = handler.read(index=self.index_name, value=value)[0]

        assert result == self.smith2


class HandlerNestingTests(BaseAuthorTestCase):

    def setUp(self):
        super(HandlerNestingTests, self).setUp()

        self.jk_name = NameAuthor.objects.create(name='JK Rowling')
        self.grisham_name = NameAuthor.objects.create(name='John Grisham')

    def test_can_nest(self):
        ahandler = Author.objects.handler()
        bhandler = NameAuthor.objects.handler()
        with ahandler, bhandler:
            handler_plain = ahandler.read()[0]
            handler_name = bhandler.read()[0]

        assert handler_plain == self.jk
        assert handler_name == self.jk_name

    def test_can_nest_two_for_same_table(self):
        ahandler = Author.objects.handler()
        bhandler = Author.objects.handler()

        with ahandler, bhandler:
            first = ahandler.read()[0]
            second = bhandler.read()[0]

        assert first == self.jk
        assert second == self.jk


class HandlerStandaloneTests(TestCase):

    def setUp(self):
        super(HandlerStandaloneTests, self).setUp()
        self.jk = VanillaAuthor.objects.create(name='JK Rowling')
        self.grisham = VanillaAuthor.objects.create(name='John Grisham')

    def test_vanilla_works(self):
        handler = Handler(VanillaAuthor.objects.all())
        with handler:
            first = handler.read()[0]

        assert first == self.jk

    def test_vanilla_filters(self):
        qs = VanillaAuthor.objects.filter(name__startswith='John')
        with Handler(qs) as handler:
            first = handler.read()[0]

        assert first == self.grisham


class HandlerMultiDBTests(TestCase):

    multi_db = True

    def setUp(self):
        super(HandlerMultiDBTests, self).setUp()
        self.jk = Author.objects.using('other').create(name='JK Rowling')
        self.grisham = Author.objects.create(name='John Grisham')

    def test_queryset_db_is_used(self):
        handler = Author.objects.using('other').handler()
        with handler:
            handler_result = handler.read()[0]

        assert handler_result == self.jk



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import json
import re

import pytest
from django import forms
from django.core import exceptions, serializers
from django.core.management import call_command
from django.db import connection, models
from django.db.migrations.writer import MigrationWriter
from django.db.models import Q
from django.test import (
    SimpleTestCase, TestCase, TransactionTestCase, override_settings,
)

from django_mysql.forms import SimpleListField
from django_mysql.models import ListCharField, ListF
from django_mysql.test.utils import override_mysql_variables
from testapp.models import (
    CharListDefaultModel, CharListModel, IntListModel, TemporaryModel,
)


class TestSaveLoad(TestCase):

    def test_char_easy(self):
        s = CharListModel.objects.create(field=["comfy", "big"])
        assert s.field == ["comfy", "big"]
        s = CharListModel.objects.get(id=s.id)
        assert s.field == ["comfy", "big"]

        s.field.append("round")
        s.save()
        assert s.field == ["comfy", "big", "round"]
        s = CharListModel.objects.get(id=s.id)
        assert s.field == ["comfy", "big", "round"]

    def test_char_string_direct(self):
        s = CharListModel.objects.create(field="big,bad")
        s = CharListModel.objects.get(id=s.id)
        assert s.field == ['big', 'bad']

    def test_is_a_list_immediately(self):
        s = CharListModel()
        assert s.field == []
        s.field.append("bold")
        s.field.append("brave")
        s.save()
        assert s.field == ["bold", "brave"]
        s = CharListModel.objects.get(id=s.id)
        assert s.field == ["bold", "brave"]

    def test_empty(self):
        s = CharListModel.objects.create()
        assert s.field == []
        s = CharListModel.objects.get(id=s.id)
        assert s.field == []

    def test_char_cant_create_lists_with_empty_string(self):
        with pytest.raises(ValueError):
            CharListModel.objects.create(field=[""])

    def test_char_cant_create_sets_with_commas(self):
        with pytest.raises(ValueError):
            CharListModel.objects.create(field=["co,mma", "contained"])

    def test_char_basic_lookup(self):
        mymodel = CharListModel.objects.create()
        empty = CharListModel.objects.filter(field="")

        assert empty.count() == 1
        assert empty[0] == mymodel

        mymodel.delete()

        assert empty.count() == 0

    def test_char_lookup_contains(self):
        self.check_char_lookup('contains')

    def test_char_lookup_icontains(self):
        self.check_char_lookup('icontains')

    def check_char_lookup(self, lookup):
        lname = 'field__' + lookup
        mymodel = CharListModel.objects.create(field=["mouldy", "rotten"])

        mouldy = CharListModel.objects.filter(**{lname: "mouldy"})
        assert mouldy.count() == 1
        assert mouldy[0] == mymodel

        rotten = CharListModel.objects.filter(**{lname: "rotten"})
        assert rotten.count() == 1
        assert rotten[0] == mymodel

        clean = CharListModel.objects.filter(**{lname: "clean"})
        assert clean.count() == 0

        with pytest.raises(ValueError):
            list(CharListModel.objects.filter(**{lname: ["a", "b"]}))

        both = CharListModel.objects.filter(
            Q(**{lname: "mouldy"}) & Q(**{lname: "rotten"}),
        )
        assert both.count() == 1
        assert both[0] == mymodel

        either = CharListModel.objects.filter(
            Q(**{lname: "mouldy"}) | Q(**{lname: "clean"}),
        )
        assert either.count() == 1

        not_clean = CharListModel.objects.exclude(**{lname: "clean"})
        assert not_clean.count() == 1

        not_mouldy = CharListModel.objects.exclude(**{lname: "mouldy"})
        assert not_mouldy.count() == 0

    def test_char_len_lookup_empty(self):
        mymodel = CharListModel.objects.create(field=[])

        empty = CharListModel.objects.filter(field__len=0)
        assert empty.count() == 1
        assert empty[0] == mymodel

        one = CharListModel.objects.filter(field__len=1)
        assert one.count() == 0

        one_or_more = CharListModel.objects.filter(field__len__gte=0)
        assert one_or_more.count() == 1

    def test_char_len_lookup(self):
        mymodel = CharListModel.objects.create(field=["red", "expensive"])

        empty = CharListModel.objects.filter(field__len=0)
        assert empty.count() == 0

        one_or_more = CharListModel.objects.filter(field__len__gte=1)
        assert one_or_more.count() == 1
        assert one_or_more[0] == mymodel

        two = CharListModel.objects.filter(field__len=2)
        assert two.count() == 1
        assert two[0] == mymodel

        three = CharListModel.objects.filter(field__len=3)
        assert three.count() == 0

    def test_char_default(self):
        mymodel = CharListDefaultModel.objects.create()
        assert mymodel.field == ["a", "d"]

        mymodel = CharListDefaultModel.objects.get(id=mymodel.id)
        assert mymodel.field == ["a", "d"]

    def test_char_position_lookup(self):
        mymodel = CharListModel.objects.create(field=["red", "blue"])

        blue0 = CharListModel.objects.filter(field__0="blue")
        assert blue0.count() == 0

        red0 = CharListModel.objects.filter(field__0="red")
        assert list(red0) == [mymodel]

        red0_red1 = CharListModel.objects.filter(field__0="red",
                                                 field__1="red")
        assert red0_red1.count() == 0

        red0_blue1 = CharListModel.objects.filter(field__0="red",
                                                  field__1="blue")
        assert list(red0_blue1) == [mymodel]

        red0_or_blue0 = CharListModel.objects.filter(
            Q(field__0="red") | Q(field__0="blue"),
        )
        assert list(red0_or_blue0) == [mymodel]

    def test_char_position_lookup_repeat_fails(self):
        """
        FIND_IN_SET returns the *first* position so repeats are not dealt with
        """
        CharListModel.objects.create(field=["red", "red", "blue"])

        red1 = CharListModel.objects.filter(field__1="red")
        assert list(red1) == []  # should be 'red'

    def test_char_position_lookup_too_long(self):
        CharListModel.objects.create(field=["red", "blue"])

        red1 = CharListModel.objects.filter(field__2="blue")
        assert list(red1) == []

    def test_int_easy(self):
        mymodel = IntListModel.objects.create(field=[1, 2])
        assert mymodel.field == [1, 2]
        mymodel = IntListModel.objects.get(id=mymodel.id)
        assert mymodel.field == [1, 2]

    def test_int_contains_lookup(self):
        onetwo = IntListModel.objects.create(field=[1, 2])

        ones = IntListModel.objects.filter(field__contains=1)
        assert ones.count() == 1
        assert ones[0] == onetwo

        twos = IntListModel.objects.filter(field__contains=2)
        assert twos.count() == 1
        assert twos[0] == onetwo

        threes = IntListModel.objects.filter(field__contains=3)
        assert threes.count() == 0

        with pytest.raises(ValueError):
            list(IntListModel.objects.filter(field__contains=[1, 2]))

        ones_and_twos = IntListModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=2),
        )
        assert ones_and_twos.count() == 1
        assert ones_and_twos[0] == onetwo

        ones_and_threes = IntListModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=3),
        )
        assert ones_and_threes.count() == 0

        ones_or_threes = IntListModel.objects.filter(
            Q(field__contains=1) | Q(field__contains=3),
        )
        assert ones_or_threes.count() == 1

        no_three = IntListModel.objects.exclude(field__contains=3)
        assert no_three.count() == 1

        no_one = IntListModel.objects.exclude(field__contains=1)
        assert no_one.count() == 0

    def test_int_position_lookup(self):
        onetwo = IntListModel.objects.create(field=[1, 2])

        one0 = IntListModel.objects.filter(field__0=1)
        assert list(one0) == [onetwo]

        two0 = IntListModel.objects.filter(field__0=2)
        assert two0.count() == 0

        one0two1 = IntListModel.objects.filter(field__0=1, field__1=2)
        assert list(one0two1) == [onetwo]


class TestListF(TestCase):

    def test_append_to_none(self):
        CharListModel.objects.create(field=[])
        CharListModel.objects.update(field=ListF('field').append('first'))
        model = CharListModel.objects.get()
        assert model.field == ["first"]

    def test_append_to_one(self):
        CharListModel.objects.create(field=["big"])
        CharListModel.objects.update(field=ListF('field').append('bad'))
        model = CharListModel.objects.get()
        assert model.field == ["big", "bad"]

    def test_append_to_some(self):
        CharListModel.objects.create(field=["big", "blue"])
        CharListModel.objects.update(field=ListF('field').append('round'))
        model = CharListModel.objects.get()
        assert model.field == ["big", "blue", "round"]

    def test_append_to_multiple_objects(self):
        CharListModel.objects.create(field=["mouse"])
        CharListModel.objects.create(field=["keyboard"])
        CharListModel.objects.update(field=ListF('field').append("screen"))
        first, second = tuple(CharListModel.objects.all())
        assert first.field == ["mouse", "screen"]
        assert second.field == ["keyboard", "screen"]

    def test_append_exists(self):
        CharListModel.objects.create(field=["nice"])
        CharListModel.objects.update(field=ListF('field').append("nice"))
        model = CharListModel.objects.get()
        assert model.field == ["nice", "nice"]

    @override_mysql_variables(SQL_MODE="ANSI")
    def test_append_works_in_ansi_mode(self):
        CharListModel.objects.create()
        CharListModel.objects.update(field=ListF('field').append('big'))
        CharListModel.objects.update(field=ListF('field').append('bad'))
        model = CharListModel.objects.get()
        assert model.field == ["big", "bad"]

    def test_append_assignment(self):
        model = CharListModel.objects.create(field=["red"])
        model.field = ListF('field').append('blue')
        model.save()
        model = CharListModel.objects.get()
        assert model.field == ['red', 'blue']

    def test_appendleft_to_none(self):
        CharListModel.objects.create(field=[])
        CharListModel.objects.update(field=ListF('field').appendleft('first'))
        model = CharListModel.objects.get()
        assert model.field == ["first"]

    def test_appendleft_to_one(self):
        CharListModel.objects.create(field=["big"])
        CharListModel.objects.update(field=ListF('field').appendleft('bad'))
        model = CharListModel.objects.get()
        assert model.field == ["bad", "big"]

    def test_appendleft_to_some(self):
        CharListModel.objects.create(field=["big", "blue"])
        CharListModel.objects.update(field=ListF('field').appendleft('round'))
        model = CharListModel.objects.get()
        assert model.field == ["round", "big", "blue"]

    def test_appendleft_to_multiple_objects(self):
        CharListModel.objects.create(field=["mouse"])
        CharListModel.objects.create(field=["keyboard"])
        CharListModel.objects.update(field=ListF('field').appendleft("screen"))
        first, second = tuple(CharListModel.objects.all())
        assert first.field == ["screen", "mouse"]
        assert second.field == ["screen", "keyboard"]

    def test_appendleft_exists(self):
        CharListModel.objects.create(field=["nice"])
        CharListModel.objects.update(field=ListF('field').appendleft("nice"))
        model = CharListModel.objects.get()
        assert model.field == ["nice", "nice"]

    @override_mysql_variables(SQL_MODE="ANSI")
    def test_appendleft_works_in_ansi_mode(self):
        CharListModel.objects.create()
        CharListModel.objects.update(field=ListF('field').appendleft('big'))
        CharListModel.objects.update(field=ListF('field').appendleft('bad'))
        model = CharListModel.objects.get()
        assert model.field == ["bad", "big"]

    def test_appendleft_assignment(self):
        model = CharListModel.objects.create(field=["red"])
        model.field = ListF('field').appendleft('blue')
        model.save()
        model = CharListModel.objects.get()
        assert model.field == ['blue', 'red']

    def test_pop_none(self):
        CharListModel.objects.create(field=[])
        CharListModel.objects.update(field=ListF('field').pop())
        model = CharListModel.objects.get()
        assert model.field == []

    def test_pop_one(self):
        CharListModel.objects.create(field=["red"])
        CharListModel.objects.update(field=ListF('field').pop())
        model = CharListModel.objects.get()
        assert model.field == []

    def test_pop_two(self):
        CharListModel.objects.create(field=["red", "blue"])
        CharListModel.objects.update(field=ListF('field').pop())
        model = CharListModel.objects.get()
        assert model.field == ["red"]

    def test_pop_three(self):
        CharListModel.objects.create(field=["green", "yellow", "p"])
        CharListModel.objects.update(field=ListF('field').pop())
        model = CharListModel.objects.get()
        assert model.field == ["green", "yellow"]

    def test_popleft_none(self):
        CharListModel.objects.create(field=[])
        CharListModel.objects.update(field=ListF('field').popleft())
        model = CharListModel.objects.get()
        assert model.field == []

    def test_popleft_one(self):
        CharListModel.objects.create(field=["red"])
        CharListModel.objects.update(field=ListF('field').popleft())
        model = CharListModel.objects.get()
        assert model.field == []

    def test_popleft_two(self):
        CharListModel.objects.create(field=["red", "blue"])
        CharListModel.objects.update(field=ListF('field').popleft())
        model = CharListModel.objects.get()
        assert model.field == ["blue"]

    def test_popleft_three(self):
        CharListModel.objects.create(field=["green", "yellow", "p"])
        CharListModel.objects.update(field=ListF('field').popleft())
        model = CharListModel.objects.get()
        assert model.field == ["yellow", "p"]


class TestValidation(SimpleTestCase):

    def test_max_length(self):
        field = ListCharField(
            models.CharField(max_length=32),
            size=3,
            max_length=32,
        )

        field.clean({'a', 'b', 'c'}, None)

        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean({'a', 'b', 'c', 'd'}, None)
        assert (
            excinfo.value.messages[0]
            == 'List contains 4 items, it should contain no more than 3.'
        )


class TestCheck(SimpleTestCase):

    def test_field_checks(self):
        class InvalidListCharModel1(TemporaryModel):
            field = ListCharField(models.CharField(), max_length=32)

        errors = InvalidListCharModel1.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E004'
        assert 'Base field for list has errors' in errors[0].msg
        assert 'max_length' in errors[0].msg

    def test_invalid_base_fields(self):
        class InvalidListCharModel2(TemporaryModel):
            field = ListCharField(
                models.ForeignKey('testapp.Author', on_delete=models.CASCADE),
                max_length=32,
            )

        errors = InvalidListCharModel2.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E005'
        assert 'Base field for list must be' in errors[0].msg

    def test_max_length_including_base(self):
        class InvalidListCharModel3(TemporaryModel):
            field = ListCharField(
                models.CharField(max_length=32),
                size=2, max_length=32)

        errors = InvalidListCharModel3.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E006'
        assert 'Field can overrun' in errors[0].msg

    def test_max_length_missing_doesnt_crash(self):
        class InvalidListCharModel4(TemporaryModel):
            field = ListCharField(
                models.CharField(max_length=2),
                size=2,
            )

        errors = InvalidListCharModel4.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'fields.E120'
        assert (
            errors[0].msg
            == "CharFields must define a 'max_length' attribute."
        )


class TestDeconstruct(TestCase):

    def test_deconstruct(self):
        field = ListCharField(models.IntegerField(), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = ListCharField(*args, **kwargs)
        assert new.base_field.__class__ == field.base_field.__class__

    def test_deconstruct_with_size(self):
        field = ListCharField(models.IntegerField(), size=3, max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = ListCharField(*args, **kwargs)
        assert new.size == field.size

    def test_deconstruct_args(self):
        field = ListCharField(models.CharField(max_length=5), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = ListCharField(*args, **kwargs)
        assert new.base_field.max_length == field.base_field.max_length


class TestMigrationWriter(TestCase):

    def test_makemigrations(self):
        field = ListCharField(models.CharField(max_length=5), max_length=32)
        statement, imports = MigrationWriter.serialize(field)

        # The order of the output max_length/size statements varies by
        # python version, hence a little regexp to match them
        assert re.compile(
            r"""^django_mysql\.models\.ListCharField\(
                models\.CharField\(max_length=5\),\ # space here
                (
                    max_length=32,\ size=None|
                    size=None,\ max_length=32
                )
                \)$
            """,
            re.VERBOSE,
        ).match(statement)

    def test_makemigrations_with_size(self):
        field = ListCharField(
            models.CharField(max_length=5),
            max_length=32,
            size=5,
        )
        statement, imports = MigrationWriter.serialize(field)

        # The order of the output max_length/size statements varies by
        # python version, hence a little regexp to match them
        assert re.compile(
            r"""^django_mysql\.models\.ListCharField\(
                models\.CharField\(max_length=5\),\ # space here
                (
                    max_length=32,\ size=5|
                    size=5,\ max_length=32
                )
                \)$
            """,
            re.VERBOSE,
        ).match(statement)


class TestMigrations(TransactionTestCase):

    @override_settings(MIGRATION_MODULES={
        "testapp": "testapp.list_default_migrations",
    })
    def test_adding_field_with_default(self):
        table_name = 'testapp_intlistdefaultmodel'
        table_names = connection.introspection.table_names
        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)

        call_command('migrate', 'testapp',
                     verbosity=0, skip_checks=True, interactive=False)
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)

        call_command('migrate', 'testapp', 'zero',
                     verbosity=0, skip_checks=True, interactive=False)
        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)


class TestSerialization(SimpleTestCase):

    def test_dumping(self):
        instance = CharListModel(field=["big", "comfy"])
        data = json.loads(serializers.serialize('json', [instance]))[0]
        field = data['fields']['field']
        assert sorted(field.split(',')) == ["big", "comfy"]

    def test_loading(self):
        test_data = '''
            [{"fields": {"field": "big,leather,comfy"},
             "model": "testapp.CharListModel", "pk": null}]
        '''
        objs = list(serializers.deserialize('json', test_data))
        instance = objs[0].object
        assert instance.field == ["big", "leather", "comfy"]


class TestDescription(SimpleTestCase):

    def test_char(self):
        field = ListCharField(models.CharField(max_length=5), max_length=32)
        assert field.description == "List of String (up to %(max_length)s)"

    def test_int(self):
        field = ListCharField(models.IntegerField(), max_length=32)
        assert field.description == "List of Integer"


class TestFormField(SimpleTestCase):

    def test_model_field_formfield(self):
        model_field = ListCharField(models.CharField(max_length=27))
        form_field = model_field.formfield()
        assert isinstance(form_field, SimpleListField)
        assert isinstance(form_field.base_field, forms.CharField)
        assert form_field.base_field.max_length == 27

    def test_model_field_formfield_size(self):
        model_field = ListCharField(models.IntegerField(), size=4)
        form_field = model_field.formfield()
        assert isinstance(form_field, SimpleListField)
        assert form_field.max_length == 4



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))
DEBUG = False
TEMPLATE_DEBUG = DEBUG

SECRET_KEY = 'THISuISdNOT9A$SECRET9x&ji!vceayg+wwt472!bgs$0!i3k4'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': 'django_mysql',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
        'OPTIONS': {
            'charset': 'utf8mb4',
            # The most forward compatible sql_mode, for 5.7
            'init_command': "SET sql_mode='STRICT_TRANS_TABLES,NO_ZERO_DATE,"
                            "NO_ZERO_IN_DATE,ERROR_FOR_DIVISION_BY_ZERO,"
                            "NO_AUTO_CREATE_USER', innodb_strict_mode=1",
        },
        'TEST': {
            'COLLATION': "utf8mb4_general_ci",
            'CHARSET': "utf8mb4",
        },
    },
    'other': {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': 'django_mysql2',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
        'OPTIONS': {
            'charset': 'utf8mb4',
            # The most forward compatible sql_mode, for 5.7
            'init_command': "SET sql_mode='STRICT_TRANS_TABLES,NO_ZERO_DATE,"
                            "NO_ZERO_IN_DATE,ERROR_FOR_DIVISION_BY_ZERO,"
                            "NO_AUTO_CREATE_USER', innodb_strict_mode=1",
        },
        'TEST': {
            'COLLATION': "utf8mb4_general_ci",
            'CHARSET': "utf8mb4",
        },
    },
    'other2': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    },
}

DATABASE_ROUTERS = [
    'testapp.routers.NothingOnSQLiteRouter',
]

ALLOWED_HOSTS = []

INSTALLED_APPS = [
    'django.contrib.admin.apps.AdminConfig',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_mysql',
    'testapp',
]

ROOT_URLCONF = 'urls'
TIME_ZONE = 'UTC'
USE_I18N = True

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
            ],
        },
    },
]


DJANGO_MYSQL_REWRITE_QUERIES = True



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import pytest
from django.db import connections
from django.test import TestCase

from django_mysql.test.utils import override_mysql_variables


class OverrideVarsMethodTest(TestCase):

    @override_mysql_variables(SQL_MODE="MSSQL")
    def test_method_sets_mssql(self):
        self.check_sql_mode("MSSQL")

    def check_sql_mode(self, expected, using='default'):
        with connections[using].cursor() as cursor:
            cursor.execute("SELECT @@SQL_MODE")
            mode = cursor.fetchone()[0]

        mode = mode.split(',')
        assert expected in mode


@override_mysql_variables(SQL_MODE="ANSI")
class OverrideVarsClassTest(OverrideVarsMethodTest):

    multi_db = True

    def test_class_sets_ansi(self):
        self.check_sql_mode("ANSI")

    @override_mysql_variables(using='other', SQL_MODE='MSSQL')
    def test_other_connection(self):
        self.check_sql_mode("ANSI")
        self.check_sql_mode("MSSQL", using='other')

    def test_it_fails_on_non_test_classes(self):
        with pytest.raises(Exception):
            @override_mysql_variables(SQL_MODE="ANSI")
            class MyClass(object):
                pass



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)


class NothingOnSQLiteRouter(object):

    def allow_migrate(self, db, app_label, **hints):
        return db in ('default', 'other')
